// ANTLR grammar parser




header {
/* ANTLR grammar parser
 * Copyright 2006 Warp IV Technologies
 * Author:  Loring Craymer
 */
}

class ANTLRBase extends Parser;
options {
	buildAST = true;
	exportVocab=ANTLRBase;
	k=2;
}

tokens {
	"tokens";
	RULE_ACTION;
	NO_ALT;
}



grammar
	:
	comments
	headerSection
	( optionsSpec )?
	( classDef )*
	EOF
	;

headerSection
	:
	(
		"header"^
		( STRING_LITERAL )?
		ACTION
		comments
	)*
	;

comments
	:
	(
		options { greedy = true; } :
		SL_COMMENT
		|
			ML_COMMENT
	)*
	;

classDef
	:
	( ACTION )?
	( DOC_COMMENT )?
	comments
	(
		("lexclass" | "class" id "extends" "Lexer" ) => lexerSpec
	|	( "class" id "extends" "TreeParser" ) => treeParserSpec
	|	parserSpec
	)
	rules
	;

optionsSpec
	:
	OPTIONS
	comments
	(
		option
	)*
	RCURLY^
	;

option
	:
		id
		ASSIGN^
		optionValue
		SEMI!
		comments
	;

lexerOptionsSpec
	:
	OPTIONS
	(
		charOption
		|
			option
	)*
	RCURLY^
	;

charOption
	:
		"charVocabulary"
		ASSIGN^
		charSet
		SEMI!
		comments
	;

optionValue
	:
	qualifiedID
	|
		STRING_LITERAL
	|
		CHAR_LITERAL
	|
		INT
	;

charSet
	:
	setBlockElement
	(
		OR^
		setBlockElement
	)*
	;

setBlockElement
	:
	CHAR_LITERAL
	(
		RANGE^
		CHAR_LITERAL
	)?
	;

tokensSpec
	:
	TOKENS^
	(
		tokenSpecItem
	)+
	RCURLY
	;

tokenSpecItem
	:
		(
			TOKEN_REF
			(
				ASSIGN^
				STRING_LITERAL
			)?
			(
				tokensSpecOptions
			)?
			|
				STRING_LITERAL
				(
					tokensSpecOptions
				)?
		)
		SEMI^
		comments
	;

tokensSpecOptions
	:
	OPEN_ELEMENT_OPTION!
	(
		option
	)+
	CLOSE_ELEMENT_OPTION^
	;

superClass
	:
	LPAREN!
	(
		STRING_LITERAL
	)
	RPAREN^
	;

parserSpec
	:
	"class"!
	id
	(
		"extends"!
		"Parser"^
		(
			superClass
		)?
	)
	SEMI!
	comments
	(
		optionsSpec
	)?
	(
		tokensSpec
	)?
	(
		ACTION
	)?
	;

lexerSpec
	:
	(
		"lexclass"^
		id
		|
			"class"!
			id
			"extends"!
			"Lexer"^
			(
				superClass
			)?
	)
	SEMI!
	comments
	(
		lexerOptionsSpec
	)?
	(
		tokensSpec
	)?
	(
		ACTION
	)?
	;

treeParserSpec
	:
	"class"!
	id
	"extends"!
	"TreeParser"^
	(
		superClass
	)?
	SEMI!
	comments
	(
		optionsSpec
	)?
	(
		tokensSpec
	)?
	(
		ACTION
	)?
	;

rules
	:
	(
		options { greedy = true; } :
		rule
	)+
	;

rule
	:
	comments
	(
		DOC_COMMENT
		comments
	)?
	(
		"protected"
		|
			"public"
		|
			"private"
	)?
	id
	(
		BANG
	)?
	(
		ARG_ACTION
	)?
	comments
	(
		"returns"
		ARG_ACTION
		comments
	)?
	(
		throwsSpec
	)?
	(
		optionsSpec
	)?
	(
		act:ACTION { #act.setType(RULE_ACTION); }
	)?
	COLON^
	block
	SEMI
	(
		exceptionGroup
	)?
	;

throwsSpec
	:
	"throws"^
	id
	(
		COMMA!
		id
	)*
	;

block
	:
	comments
	alternative
	(
		OR
		comments
		alternative
	)*
	;

alternative
	:
	(
		b:BANG { #b.setType(NO_ALT); }
	)?
	comments
	(
		element
	)*
	(
		exceptionSpecNoLabel
	)?
	;

exceptionGroup
	:
	(
		exceptionSpec
	)+
	;

exceptionSpec
	:
	"exception"^
	(
		ARG_ACTION
	)?
	(
		exceptionHandler
	)*
	;

exceptionSpecNoLabel
	:
	"exception"^
	(
		exceptionHandler
	)*
	;

exceptionHandler
	:
	"catch"^
	ARG_ACTION
	ACTION
	;

element
	:
	elementNoOptionSpec
	comments
	(
		elementOptionSpec
	)?
	;

elementOptionSpec
	:
	OPEN_ELEMENT_OPTION!
	(
		option
	)+
	CLOSE_ELEMENT_OPTION^
	;

elementNoOptionSpec
	:
	id
	ASSIGN^
	(
		id
		COLON^
	)?
	(
		RULE_REF
		(
			ARG_ACTION
		)?
		(
			BANG
		)?
		|
			TOKEN_REF
			(
				ARG_ACTION
			)?
	)
	comments
	|
		(
			id
			COLON^
		)?
		(
			RULE_REF
			(
				ARG_ACTION
			)?
			(
				BANG
			)?
			|
				range
			|
				terminal
			|
				NOT_OP
				(
					notTerminal
					|
						ebnf
				)
			|
				ebnf
		)
		comments
	|
		ACTION
		comments
	|
		SEMPRED
		comments
	|
		tree
		comments
	;

tree
	:
	TREE_BEGIN^
	rootNode
	(
		element
	)+
	RPAREN!
	;

rootNode
	:
	(
		id
		COLON^
	)?
	terminal
	comments
	;


ebnf
	:
	LPAREN^
	comments
	(
		optionsSpec
		(
			ACTION
		)?
		COLON
		|
			( ACTION COLON )=>
			ACTION
			COLON
	)?
	block
	RPAREN
	(
		(
			QUESTION
			|
				STAR
			|
				PLUS
		)?
		(
			BANG
		)?
		|
			IMPLIES
	)
	comments
	;

ast_type_spec
	:
	(
		CARET
		|
			BANG
	)?
	;

range
	:
	CHAR_LITERAL
	RANGE^
	CHAR_LITERAL
	(
		BANG
	)?
	|
		(
			TOKEN_REF
			|
				STRING_LITERAL
		)
		RANGE^
		(
			TOKEN_REF
			|
				STRING_LITERAL
		)
		ast_type_spec
	;

terminal
	:
	CHAR_LITERAL
	(
		BANG
	)?
	|
		TOKEN_REF
		ast_type_spec
		(
			ARG_ACTION
		)?
	|
		STRING_LITERAL
		ast_type_spec
	|
		WILDCARD
		ast_type_spec
	;

notTerminal
	:
	CHAR_LITERAL
	(
		BANG
	)?
	|
		TOKEN_REF
		ast_type_spec
	;

qualifiedID
	:
	id
	(
		WILDCARD^
		id
	)*
	;

id
	:
	TOKEN_REF
	|
		RULE_REF
	;

// ANTLRLexer licensing:
/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.cs.usfca.edu
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.6/antlr/antlr.g#1 $
 */

class ANTLRLexer extends Lexer;
options {
	k=2;
	exportVocab=ANTLRBase;
	testLiterals=false;
	interactive=true;
	charVocabulary='\003'..'\377';
}

tokens {
	"options";
}

{
	/**Convert 'c' to an integer char value. */
	public static int escapeCharValue(String cs) {
		//System.out.println("escapeCharValue("+cs+")");
		if ( cs.charAt(1)!='\\' ) return 0;
		switch ( cs.charAt(2) ) {
		case 'b' : return '\b';
		case 'r' : return '\r';
		case 't' : return '\t';
		case 'n' : return '\n';
		case 'f' : return '\f';
		case '"' : return '\"';
		case '\'' :return '\'';
		case '\\' :return '\\';

		case 'u' :
			// Unicode char
			if (cs.length() != 8) {
				return 0;
			}
			else {
				return
					Character.digit(cs.charAt(3), 16) * 16 * 16 * 16 +
					Character.digit(cs.charAt(4), 16) * 16 * 16 +
					Character.digit(cs.charAt(5), 16) * 16 +
					Character.digit(cs.charAt(6), 16);
			}

		case '0' :
		case '1' :
		case '2' :
		case '3' :
			if ( cs.length()>5 && Character.isDigit(cs.charAt(4)) ) {
				return (cs.charAt(2)-'0')*8*8 + (cs.charAt(3)-'0')*8 + (cs.charAt(4)-'0');
			}
			if ( cs.length()>4 && Character.isDigit(cs.charAt(3)) ) {
				return (cs.charAt(2)-'0')*8 + (cs.charAt(3)-'0');
			}
			return cs.charAt(2)-'0';

		case '4' :
		case '5' :
		case '6' :
		case '7' :
			if ( cs.length()>4 && Character.isDigit(cs.charAt(3)) ) {
				return (cs.charAt(2)-'0')*8 + (cs.charAt(3)-'0');
			}
			return cs.charAt(2)-'0';

		default :
			return 0;
		}
	}

	public static int tokenTypeForCharLiteral(String lit) {
		if ( lit.length()>3 ) {  // does char contain escape?
			return escapeCharValue(lit);
		}
		else {
			return lit.charAt(1);
		}
	}
}

WS	:	(	/*	'\r' '\n' can be matched in one alternative or by matching
				'\r' in one iteration and '\n' in another.  I am trying to
				handle any flavor of newline that comes in, but the language
				that allows both "\r\n" and "\r" and "\n" to all be valid
				newline is ambiguous.  Consequently, the resulting grammar
				must be ambiguous.  I'm shutting this warning off.
			 */
			options {
				generateAmbigWarnings=false;
			}
		:	' '
		|	'\t'
		|	'\r' '\n'	{newline();}
		|	'\r'		{newline();}
		|	'\n'		{newline();}
		)
		{ $setType(Token.SKIP); }
	;

COMMENT :
	(
		sl:SL_COMMENT
		{$setType(sl.getType());}
		|
			t:ML_COMMENT
			{$setType(t.getType());}
	)
	;

protected
SL_COMMENT :
	"//"
	( ~('\n'|'\r') )*
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then in the next token.  The language
			that allows both "\r\n" and "\r" and "\n" to all be valid
			newline is ambiguous.  Consequently, the resulting grammar
			must be ambiguous.  I'm shutting this warning off.
		 */
			options {
				generateAmbigWarnings=false;
			}
		:	'\r' '\n'
		|	'\r'
		|	'\n'
	)
	{ newline(); }
	;

protected
ML_COMMENT :
	"/*"
	(	{ LA(2)!='/' }? '*' {$setType(DOC_COMMENT);}
	|
	)
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then in the next token.  The language
			that allows both "\r\n" and "\r" and "\n" to all be valid
			newline is ambiguous.  Consequently, the resulting grammar
			must be ambiguous.  I'm shutting this warning off.
		 */
		options {
			greedy=false;  // make it exit upon "*/"
			generateAmbigWarnings=false; // shut off newline errors
		}
	:	'\r' '\n'	{newline();}
	|	'\r'		{newline();}
	|	'\n'		{newline();}
	|	~('\n'|'\r')
	)*
	"*/"
	;

OPEN_ELEMENT_OPTION
	:	'<'
	;

CLOSE_ELEMENT_OPTION
	:	'>'
	;

COMMA : ',';

QUESTION :	'?' ;

TREE_BEGIN : "#(" ;

LPAREN:	'(' ;

RPAREN:	')' ;

COLON :	':' ;

STAR:	'*' ;

PLUS:	'+' ;

ASSIGN : '=' ;

IMPLIES : "=>" ;

SEMI:	';' ;

CARET : '^' ;

BANG : '!' ;

OR	:	'|' ;

WILDCARD : '.' ;

RANGE : ".." ;

NOT_OP :	'~' ;

RCURLY:	'}'	;

CHAR_LITERAL
	:	'\'' (ESC|~'\'') '\''
	;

STRING_LITERAL
	:	'"' (ESC|~'"')* '"'
	;

protected
ESC	:	'\\'
		(	'n'
		|	'r'
		|	't'
		|	'b'
		|	'f'
		|	'w'
		|	'a'
		|	'"'
		|	'\''
		|	'\\'
		|	('0'..'3')
			( options { warnWhenFollowAmbig = false; }: '0'..'7'
			( options { warnWhenFollowAmbig = false; }: '0'..'7'
		   )?
			)?
		|	('4'..'7')
			( options { warnWhenFollowAmbig = false; }: '0'..'7' )?
		|	'u' XDIGIT XDIGIT XDIGIT XDIGIT
		)
	;

protected
DIGIT
	:	'0'..'9'
	;

protected
XDIGIT :
		'0' .. '9'
	|	'a' .. 'f'
	|	'A' .. 'F'
	;

INT	:	('0'..'9')+
	;

ARG_ACTION
   :
	NESTED_ARG_ACTION
	;

protected
NESTED_ARG_ACTION :
	'['
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then '\n' in the next iteration.
		 */
		options {
			generateAmbigWarnings=false; // shut off newline errors
		}
	:	NESTED_ARG_ACTION
	|	'\r' '\n'	{newline();}
	|	'\r'		{newline();}
	|	'\n'		{newline();}
	|	CHAR_LITERAL
	|	STRING_LITERAL
	|	~']'
	)*
	']'
	;

ACTION
{int actionLine=getLine(); int actionColumn = getColumn(); }
	:	NESTED_ACTION
		(	'?'	{_ttype = SEMPRED;} )?
		{
			CommonToken t = new CommonToken(_ttype,$getText);
			t.setLine(actionLine);			// set action line to start
			t.setColumn(actionColumn);
			$setToken(t);
		}
	;

protected
NESTED_ACTION :
	'{'
	(
		options {
			greedy = false; // exit upon '}'
		}
	:
		(
			options {
				generateAmbigWarnings = false; // shut off newline warning
			}
		:	'\r' '\n'	{newline();}
		|	'\r' 		{newline();}
		|	'\n'		{newline();}
		)
	|	NESTED_ACTION
	|	CHAR_LITERAL
	|	COMMENT
	|	STRING_LITERAL
	|	.
	)*
	'}'
   ;

TOKEN_REF
options { testLiterals = true; }
	:	'A'..'Z'
		(	// scarf as many letters/numbers as you can
			options {
				warnWhenFollowAmbig=false;
			}
		:
			'a'..'z'|'A'..'Z'|'_'|'0'..'9'
		)*
	;

// we get a warning here when looking for options '{', but it works right
RULE_REF
{
	int t=0;
}
	:	t=INTERNAL_RULE_REF {_ttype=t;}
		(	{t==LITERAL_options}? WS_LOOP ('{' {_ttype = OPTIONS;})?
		|	{t==LITERAL_tokens}? WS_LOOP ('{' {_ttype = TOKENS;})?
		|
		)
	;

protected
WS_LOOP
	:	(	// grab as much WS as you can
			options {
				greedy=true;
			}
		:
			WS
		|	COMMENT
		)*
	;

protected
INTERNAL_RULE_REF returns [int t]
{
	t = RULE_REF;
}
	:	'a'..'z'
		(	// scarf as many letters/numbers as you can
			options {
				warnWhenFollowAmbig=false;
			}
		:
			'a'..'z'|'A'..'Z'|'_'|'0'..'9'
		)*
		{t = testLiteralsTable(t);}
	;

protected
WS_OPT :
	(WS)?
	;
